import os.path


def find_sensitive_path(dir, insensitive_path):
    insensitive_path = insensitive_path.strip(os.path.sep)

    parts = insensitive_path.split(os.path.sep)
    next_name = parts[0]
    for name in os.listdir(dir):
        if next_name.lower() == name.lower():
            improved_path = os.path.join(dir, name)
            if len(parts) == 1:
                return improved_path
            else:
                return find_sensitive_path(improved_path, os.path.sep.join(parts[1:]))
    return None




"""
Fury of the Furries level loader for Tiled
2012, <samuli@tuomola.net>
"""

import sys, re
import tiled as T
from os.path import dirname
from lib import cpystruct, lbm, utils
from struct import pack,unpack
from collections import namedtuple

maps = {}
tilesets = {}

MetaData = namedtuple(
  'MetaData', 'startX, startY, lType,'+
  'fin1X, fin1Y, u1, fin2X, fin2Y, lvl'
  )

class Fury(T.Plugin):
  @classmethod
  def nameFilter(cls):
    return "Fury of the Furries (*.bin)"

  @classmethod
  def supportsFile(cls, f):
    return open(f,'rb').read(4) == b'byt4'

  @classmethod
  def read(cls, f):
    print('Loading map at',f)
    fr = Fury(f)

    m = T.Tiled.Map(T.Tiled.Map.Orthogonal, fr.w, fr.h, 16, 16)
    maps[f] = m

    # probably defined explicitly somewhere in the data
    decs = [1,3,4,2,8,6,7,9,10,5]
    decnum = (int(re.findall('[0-9]+', f).pop())-1)/10
    if decnum >= len(decs): decnum %= len(decs)
    gfxf = 'DEC/DECOR%02i.LBM' % decs[int(decnum)]
    gfxf = utils.find_sensitive_path(dirname(f)+'/../', gfxf)
    t = T.Tiled.Tileset.create('DECOR', 16,16, 0, 0)
    t.data().loadFromImage(fr.readtilegfx(gfxf), '')
    l = T.Tiled.TileLayer('Tiles',0,0, fr.w, fr.h)
    fr.populatetiles(l, t.data())
    # have to pass ownership so can't add tileset before populating layer
    m.addTileset(t)
    m.addLayer(l)

    #if not f.endswith('DATA19.BIN'):
    #  del fr.lvl[:21*2] # skip some zeros, todo: more deterministic
    print( len(fr.lvl), unpack('<40h', fr.lvl[:80]) )
    print( MetaData(*unpack('<9h', fr.lvl[:18])) )

    return m

  @classmethod
  def write(cls, m, fn):
    print("-- script doesn't support writing yet")
    fn += '.testing'
    print(".. map(%i,%i) to" % (m.width(),m.height()), fn)
    lvl = []

    with open(fn, 'w') as fh:
      for i in range(m.layerCount()):
        tiles = []
        l = 0
        if isTileLayerAt(m, i):
          l = tileLayerAt(m, i)
          print(l)
        elif isObjectGroupAt(m, i):
          #l = objectGroupAt(m, i)
          continue

        for x in range(l.width()):
          for y in range(l.height()):
            tiles.append( l.cellAt(x, y).tile.id() )

        print(tiles, file=fh)

    return False

  def __init__(self, f):
    dat = bytearray()
    with open(f, 'rb') as fh:
      fh.read(4)  #skip sig
      ldata = LevelData()
      rdata = RleData()
      while ldata.unpack(fh) and ldata.len != 0 :
        if len(ldata.d)==1 and ldata.d[0] == 0: break
        rdata.unpack(fh)
        dat.extend(ldata.d)
        dat.extend([rdata.val for i in range(rdata.rep)])

    print('le',len(dat))
    self.w, self.h = unpack("<2H", dat[:4])
    del dat[:4]
    self.lvl = dat

  def readtilegfx(self, fn):
    lc = dict(lbm.parselbm(fn))
    #print(lc['BMHD'])
    bd = list(lbm.readbody(lc['BODY'], lc['BMHD']))
    img = T.qt.QImage(lc['BMHD'].sz.w, lc['BMHD'].sz.h, T.qt.QImage.Format_Indexed8)
    img.setColorTable(lc['CMAP'])

    # there should be a faster alternative
    for y in range(img.height()):
      for x in range(img.width()):
        img.setPixel(x, y, bd[y*img.width()+x])

    return img

  def populatetiles(self, l, t):
    i = 0
    for y in range(self.h):
      for x in range(self.w):
        tpos = self.lvl[i+1]*t.columnCount() + self.lvl[i]
        if tpos < t.tileCount():
          ti = t.tileAt(tpos)
          if ti != None:
            l.setCell(x, y, T.Tiled.Cell(ti))
        i += 2
      if self.w < 78: # padded to width of 78 tiles
        i += (78-self.w)*2

    del self.lvl[:self.w*self.h*2]

class LevelData(cpystruct.CpyStruct('ushort len; BYTE d[len];')): pass

class RleData(cpystruct.CpyStruct('BYTE rep, sig, val')): pass




"""
  ILBM pure python decoder
  2012, <samuli@tuomola.net>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import tiled as T
from .cpystruct import *
#from PyQt4.Qt import QColor  # to test outside tiled
import struct

class IFFhead(CpyStruct(":4s id; :I len; :4s type;", True)): pass

class IFFchunk(CpyStruct(":4s id; :I len;", True)):
  @classmethod
  def parsefile(cls, f):
    """
    Reads whole file consisting of IFF chunks
    Returns tuple generator of chunk name and body
    """
    with open(f, 'rb') as fh:
      # header
      ih = IFFhead(fh)
      yield ih.type, ih
      # chunks
      ic = IFFchunk()
      filelen = ih.len - len(ih)
      while fh.tell() < filelen and ic.unpack(fh) != None:
        yield ic.id, fh.read(ic.len)

class BMHDsize(CpyStruct("short w, h", True)):
  @classmethod
  def fromraw(cls, v):
    return struct.unpack(getattr(cls,'__fstr'), v)

class BMHD(CpyStruct("""
  BMHDsize sz; short x, y;
  BYTE planes, mask; short comp, trans;
  BYTE xaspect, yaspect; short pgw, pgh;
""")): pass

class CRNG(CpyStruct("short pad, cycrate, cycle, lowhireg;", True)): pass

class CMAP(CpyStruct("BYTE color[3];", True)):
  @classmethod
  def parse(cls, dat):
    c = CMAP()
    sz = struct.calcsize(getattr(cls, '__fstr'))
    for i in range(int(len(dat)/sz)):
      c.unpack(dat[i*sz:i*sz+sz])
      yield T.qt.QColor(c.color[0],c.color[1],c.color[2]).rgb()

def uncomp(dat):
  i = 0
  ret = bytearray()
  while i < len(dat):
    v = struct.unpack('b', dat[i:i+1])[0]
    i += 1
    if v >= 0:
      for n in range(v+1):
        ret += dat[i:i+1]
        i += 1
    elif v != -128:
      for n in range(v-1,0,1):
        ret += dat[i:i+1]
      i += 1
  return ret

def readbody(bd, ch):
  bj = [0,int((ch.sz.w+15) / 16) * 2]
  for p in range(2,ch.planes):
    bj.append(bj[1] * p)

  sj = bj[1] * ch.planes
  for sl in range(0, sj*ch.sz.h, sj):
    for x in range(ch.sz.w):
      b = x & 7
      bm = 128 >> b
      ib = sl + (x >> 3)
      px = (bd[ib] & bm)
      # only 4 planes tested
      for p in range(1,ch.planes,1):
        px |= (bd[ib+bj[p]] & bm) << p
      yield px >> (7 - b)

def parselbm(f):
  for id,dat in list(IFFchunk.parsefile(f)):
    if id == 'BMHD':
      ch = BMHD(dat)
      yield id, ch
    elif id == 'CMAP':
      yield id, list(CMAP.parse(dat))
    elif id == 'CRNG':
      yield id, CRNG(dat)
    elif id == 'CAMG':
      raise Exception('HAM/HALFBRITE not supported at the moment')
    elif id == 'BODY':
      if ch.comp == 1:
        bd = uncomp(dat)
      else:
        bd = dat
      yield id, bd

if __name__ == "__main__":
  import sys
  print( list(parselbm(sys.argv[1])) )




"""
Pekka Kana 2 map support for Tiled
2012, <samuli@tuomola.net>

Notes:
- should make PK2 classes mixins with Tiled ones
"""
from lib import cpystruct
from os.path import dirname, exists
from struct import pack,unpack,Struct
from base64 import b64encode, b64decode
import os, sys, re, string
import tiled as T

maps = []

class PK2(T.Plugin):
  @classmethod
  def nameFilter(cls):
    return "Pekka Kana 2 (*.map)"

  @classmethod
  def shortName(cls):
    return "pk2"

  @classmethod
  def supportsFile(cls, f):
    return open(f, 'rb').read(4) == b'1.3\0'

  @classmethod
  def read(cls, f):
    lvl = PK2MAP()
    with open(f, 'rb') as fh:
      lvl.unpack(fh)
      # spriteCount is +1
      fh.seek(-len(lvl.spriteFiles[0]), 1)
      lay1 = PK2MAPLAYER(fh)
      lay2 = PK2MAPLAYER(fh)
      lay3 = PK2MAPLAYER(fh)
      print(lvl)

    # -- tileset
    img = T.qt.QImage()
    imgfile = dirname(f)+'/../../gfx/tiles/'+str(lvl.tileFile)
    img.load(imgfile, 'BMP')
    t = T.Tiled.Tileset.create('Tiles', 32,32, 0, 0)
    t.data().setTransparentColor(T.qt.QColor(img.color(255)))
    t.data().loadFromImage(img, imgfile)

    # find common bounding box for the layers
    bb = ['','',10,10]
    for l in [lay1,lay2,lay3]:
      print(l)
      bb[0] = min([bb[0], l.lx.num])
      bb[1] = min([bb[1], l.ly.num])
      bb[2] = max([bb[2], l.width()])
      bb[3] = max([bb[3], l.height()])

    print('bounds', bb)

    m = T.Tiled.Map(T.Tiled.Map.Orthogonal, bb[2], bb[3], 32,32)
    maps.append(m)

    # -- background image
    lai = T.Tiled.ImageLayer('Scenery', bb[2], bb[3])
    img = T.qt.QImage()
    imgfile = dirname(f)+'/../../gfx/scenery/'+str(lvl.fieldFile)
    img.load(imgfile, 'BMP')
    lai.loadFromImage(img, imgfile)

    # -- layers
    la1 = T.Tiled.TileLayer('Back', 0,0, bb[2], bb[3])
    lay1.doTiles(t, la1, bb)

    la2 = T.Tiled.TileLayer('Front', 0,0, bb[2], bb[3])
    lay2.doTiles(t, la2, bb)

    sprdir = dirname(f)+'/../../sprites/'
    lay3.sprites = [0]
    lay3.spriteGfx = {}

    for s in lvl.spriteFiles:
      # spriteCount is +1
      if not exists(sprdir+str(s)): break

      spr = PK2SPR(sprdir+str(s), m)

      if not lay3.spriteGfx.has_key(str(spr.kuvatiedosto)):
        sprfile = find_case_insensitive_filename(sprdir, str(spr.kuvatiedosto))
        img = T.qt.QImage()
        img.load(sprdir+sprfile, 'BMP')
        print('loading', sprdir+sprfile)
        sprts = T.Tiled.Tileset.create(sprfile, 32,32, 0, 0)
        sprts.data().setTransparentColor(T.qt.QColor(img.color(255)))
        sprts.data().loadFromImage(img, sprdir+sprfile)
        lay3.spriteGfx[str(spr.kuvatiedosto)] = sprts

      #sprgfx[(str(spr.kuvatiedosto]
      lay3.sprites.append(spr)

      #print spr

    la3 = T.Tiled.ObjectGroup('Sprites', bb[2], bb[3])
    lay3.doSprites(la3, bb)
    m.addLayer(lai)
    m.addTileset(t)
    for sprts in lay3.spriteGfx.values():
      m.addTileset(sprts)
    m.addLayer(la1)
    m.addLayer(la2)
    m.addLayer(la3)

    for f in lvl.__slots__:
      val = repr(getattr(lvl, f))
      m.setProperty(f, b64encode(val))

    return m

  @classmethod
  def write(cls, m, fn):
    out = PK2MAP()

    for f in m.properties().keys():
      if not f.startswith('__'):
        setattr(out, f, b64decode(m.property(f)))

    #setattr(out, "sprites", ['pla'])

    with open(fn, 'wb') as fh:
      print(out.pack(), file=fh)

      for i in range(m.layerCount()):
        tiles = []
        l = 0
        if isTileLayerAt(m, i):
          l = tileLayerAt(m, i)
          print(l)
        elif isObjectGroupAt(m, i):
          #l = objectGroupAt(m, i)
          continue

        for y in range(l.height()):
          for x in range(l.width()):
            if l.cellAt(x, y).tile != None:
              tiles.append( l.cellAt(x, y).tile.id() )
        print(0,0, l.width(), l.height(), file=fh)
        print(bytearray(tiles), file=fh)

    return True

def find_case_insensitive_filename(path, fn):
  for f in os.listdir(path):
    if f.lower() == fn.lower():
      return f

class asciilongfile(cpystruct.CpyStruct("char filename[100]")):
  @classmethod
  def fromraw(cls, v):
    return re.search('[\w\.]*', v, re.U).group(0)
  def __repr__(self):
    return str(self)
  def __str__(self):
    return self.filename

class asciifile(cpystruct.CpyStruct("char filename[13]")):
  @classmethod
  def fromraw(cls, v):
    return re.search('[\w\.]*', v, re.U).group(0)
  def __repr__(self):
    return str(self)
  def __str__(self):
    return self.filename

class asciitxt(cpystruct.CpyStruct("char txt[40]")):
  @classmethod
  def fromraw(cls, v):
    return re.search('[\w\. ]', v, re.U).group(0)

class asciinum(cpystruct.CpyStruct("char num[8]")):
  @classmethod
  def fromraw(cls, v):
    #v = ''.join(re.findall('[0-9]', v))
    v = re.sub('[^0-9]','',v)
    return 0 if not v.strip().isdigit() else int(v)

class PK2MAPLAYER(cpystruct.CpyStruct("asciinum lx, ly, w, h;")):
  MAXW = 256
  MAXH = 224
  MAXSZ = MAXW*MAXH

  def width(self):
    return self.w.num + 1
  def height(self):
    return self.h.num + 1

  def __init__(self, dat):
    # should make cpystruct support this usecase better
    super(self.__class__, self).__init__(dat)

    #print str(cpystruct.peek(dat, 128))

    self.layer = bytearray(self.MAXSZ)
    for i in range(len(self.layer)): self.layer[i] = 0xff

    for y in range(self.ly.num, self.ly.num+self.height()):
      for x in range(self.lx.num, self.lx.num+self.width()):
        self.layer[x+y*self.MAXW] = dat.read(1)

  def findBounds(self):
    "find bounding box for coords that have tiles"
    mx,my,mw,mh = None,None,10,10

    for y in range(self.ly, self.ly+self.height()):
      for x in range(self.lx, self.lx+self.width()):
        if self.layer[x + y * self.MAXW] != 255:
          if not my: my = y
          if not mx or x < mx: mx = x
          if x > mw: mw = x
          if y > mh: mh = y

    if not mx: mx = 0
    if not my: my = 0
    return mx, my, mw, mh

  def doSprites(self, la, bb):
    for y in range(self.height()):
      for x in range(self.width()):
        sprite = self.layer[self.lx.num + x + (self.ly.num + y) * self.MAXW]
        if sprite != 255:
          #if sprite > len(self.sprites):
          #  print 'invalid spr',sprite
          #  continue
          rx = self.lx.num + x - bb[0]
          ry = self.ly.num + y - bb[1] + 1
          spr = self.sprites[sprite]
          obj = T.Tiled.MapObject(str(spr.kuvatiedosto), '', T.qt.QPointF(rx, ry), T.qt.QSizeF(1, 1)) #spr.width, spr.height))
          # 0 should point to the actual sprite but how?
          obj.setCell(Tiled.Cell(self.spriteGfx[str(spr.kuvatiedosto)].data().tileAt(0)))
          la.addObject(obj)

  def doTiles(self, ts, la, bb):
    for y in range(self.height()):
      for x in range(self.width()):
        tile = self.layer[self.lx.num + x + (self.ly.num + y) * self.MAXW]
        if tile != 255:
          rx = self.lx.num + x - bb[0]
          ry = self.ly.num + y - bb[1]
          if tile == 149: print('start @',rx,ry)
          if tile == 150: print('end @',rx,ry)
          ti = ts.data().tileAt(tile)
          if ti != None and rx < bb[2] and ry < bb[3]:
            # app should check that coords are within layer
            #print rx,ry,self.ly,y
            la.setCell(rx, ry, T.Tiled.Cell(ti))
          else:
            print('invalid',rx,ry)

class PK2SPR_ANIM(cpystruct.CpyStruct("""
  uchar   seq[10];
  uchar   frames;
  bool    loop;
""")):
  def __repr__(self):
    return str(self)

class PK2SPR(cpystruct.CpyStruct("""
asciinum  tyyppi;
asciilongfile  kuvatiedosto;
asciilongfile  aanitiedostot[7];
int     aanet[7];
uchar   frameja;
PK2SPR_ANIM animaatiot[20];
uchar   animaatioita;
uchar   frame_rate;
int     kuva_x;
int     kuva_y;
int     kuva_frame_leveys;
int     kuva_frame_korkeus;
int     kuva_frame_vali;
char    nimi[30];
int     width, height;
double  paino;
bool    vihollinen;
int     energia;
int     vahinko;
uchar   vahinko_tyyppi;
uchar   suojaus;
int     pisteet;
int     AI[10];
uchar   max_hyppy;
double  max_nopeus;
int     latausaika;
uchar   vari;
bool    este;
int     tuhoutuminen;
bool    avain;
bool    tarisee;
uchar   bonusten_lkm;
int     hyokkays1_aika;
int     hyokkays2_aika;
int     pallarx_kerroin;
char    muutos_sprite[100];
char    bonus_sprite[100];
char    ammus1_sprite[100];
char    ammus2_sprite[100];

bool    tiletarkistus;
DWORD   aani_frq;
bool    random_frq;

bool    este_ylos;
bool    este_alas;
bool    este_oikealle;
bool    este_vasemmalle;

uchar   lapinakyvyys;
bool    hehkuu;
int     tulitauko;
bool    liitokyky;
bool    boss;
bool    bonus_aina;
bool    osaa_uida;
""")):
  def __init__(self, f, m):
    with open(f, 'rb') as fh:
      super(self.__class__, self).__init__(fh)

class PK2MAP(cpystruct.CpyStruct("""
  char ver[4];
  BYTE nan;
  asciifile tileFile, fieldFile, musicFile;
  asciitxt mapName, author;
  asciinum lvl, air, trig1, trig2, trig3;
  asciinum playTime, v1, background, plrSpr;
  asciinum lvlX, lvlY, icon;
  asciinum spriteCount;
  asciifile spriteFiles[spriteCount];
""")): pass




"""
 Python Tiled Plugin
 Copyright 2012, Samuli Tuomola <samuli@tuomola.net>

 This file is part of Tiled.

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the Free
 Software Foundation; either version 2 of the License, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 more details.

 You should have received a copy of the GNU General Public License along with
 this program. If not, see <http://www.gnu.org/licenses/>.
"""

import re
from pybindgen import *
import pybindgen.typehandlers.base as typehandlers
from pybindgen.typehandlers.base import ForwardWrapperBase, PointerParameter

"""
class QFlagsTransformation(typehandlers.TypeTransformation):
  def __init__(self):
    self.rx = re.compile(r'(?:::)?QFlags<\s*(\w+)\s*>')

  def get_untransformed_name(self, name):
    m = self.rx.match(name)
    if m is None:
      return None
    else:
      return m.group(1)+' *'

  def create_type_handler(self, type_handler, *args, **kwargs):
    ctype = self.get_untransformed_name(args[0])
    handler = type_handler(ctype, *args[1:], **kwargs)
    handler.has_been_transformed = True
    return handler

typehandlers.param_type_matcher.register_transformation(QFlagsTransformation())
"""

class QFlagsOptionParam(Parameter):
  DIRECTIONS = [Parameter.DIRECTION_IN]
  CTYPES = ['QFlags<QFileDialog::Option>']

  def get_c_error_return(self):
    return "return QFlags(0);"

  def convert_c_to_python(self, wrapper):
    raise NotImplementedError
    #wrapper.build_params.add_parameter("s", ['%s.toUtf8().data()' % self.value], prepend=True)

  def convert_python_to_c(self, wrapper):
    name = wrapper.declarations.declare_variable("QFileDialog::Option", self.name)
    wrapper.parse_params.add_parameter('i', ['&'+name], self.value, optional=bool(self.default_value))
    if self.default_value is None:
      wrapper.call_params.append('(QFlags<QFileDialog::Option>)%s' % name)
    else:
      wrapper.call_params.append(self.default_value)

class QStringPtrParam(PointerParameter):
  DIRECTIONS = [Parameter.DIRECTION_IN]	# could be out as well
  CTYPES = ['QString*']

  def convert_c_to_python(self, wrapper):
    raise NotImplementedError

  def convert_python_to_c(self, wrapper):
    name = wrapper.declarations.declare_variable("const char *", self.name)
    if self.default_value is None:
      name_qst = wrapper.declarations.declare_variable("QString*", self.name + '_qst',
        'new QString(%s)' % name)
      wrapper.call_params.append('%s' % name_qst)
    else:
      wrapper.call_params.append(self.default_value)
    wrapper.parse_params.add_parameter('s', ['&'+name], self.value, optional=bool(self.default_value))

class QStringParam(Parameter):
  DIRECTIONS = [Parameter.DIRECTION_IN]
  CTYPES = ['QString']

  def get_c_error_return(self):
    return "return QString();"

  def convert_c_to_python(self, wrapper):
    wrapper.build_params.add_parameter("s", ['%s.toUtf8().data()' % self.value], prepend=True)

  def convert_python_to_c(self, wrapper):
    name = wrapper.declarations.declare_variable("const char *", self.name)
    len_ = wrapper.declarations.declare_variable("Py_ssize_t", self.name+"_len")
    wrapper.parse_params.add_parameter('s#', ['&'+name, '&'+len_], self.value)
    wrapper.call_params.append('QString::fromUtf8(%s)' % name)

class QStringReturnValue(ReturnValue):
  CTYPES = ['QString']

  def get_c_error_return(self):
    return "return QString();"

  def convert_python_to_c(self, wrapper):
    #raise NotImplementedError # TODO (needed only for virtual methods where C calls Python code)
    ptr = wrapper.declarations.declare_variable("const char *", "retval_ptr")
    len_ = wrapper.declarations.declare_variable("Py_ssize_t", "retval_len")
    wrapper.parse_params.add_parameter("s#", ['&'+ptr, '&'+len_])
    wrapper.after_call.write_code("%s = QString(%s);" % (self.value, ptr))

  def convert_c_to_python(self, wrapper):
    wrapper.build_params.add_parameter("s", ['%s.toUtf8().data()' % self.value], prepend=True)

def generate(parent_mod):
  mod = module.SubModule('qt', parent_mod)
  mod.add_include('<QImage>')
  mod.add_include('<QFileDialog>')
  mod.add_include('<QWidget>')
  mod.add_include('<QFlags>')
  #mod.add_include('"qtbind.h"')

  cls_qpoint = mod.add_class('QPoint')
  cls_qpoint.add_copy_constructor()
  cls_qpoint.add_constructor([('int','x'),('int','y')])
  cls_qpoint.add_method('x', 'int', [])
  cls_qpoint.add_method('setX', None, [('int','x')])
  cls_qpoint.add_method('y', 'int', [])
  cls_qpoint.add_method('setY', None, [('int','y')])

  cls_qpointf = mod.add_class('QPointF')
  cls_qpointf.add_copy_constructor()
  cls_qpointf.add_constructor([('double','x'),('double','y')])
  cls_qpointf.add_method('x', 'double', [])
  cls_qpointf.add_method('setX', None, [('double','x')])
  cls_qpointf.add_method('y', 'double', [])
  cls_qpointf.add_method('setY', None, [('double','y')])

  cls_size = mod.add_class('QSize')
  cls_size.add_copy_constructor()
  cls_size.add_constructor([('int','w'),('int','h')])
  cls_size.add_method('width', 'int', [])
  cls_size.add_method('setWidth', None, [('int','w')])
  cls_size.add_method('height', 'int', [])
  cls_size.add_method('setHeight', None, [('int','h')])

  cls_sizef = mod.add_class('QSizeF')
  cls_sizef.add_copy_constructor()
  cls_sizef.add_constructor([('double','w'),('double','h')])
  cls_sizef.add_method('width', 'double', [])
  cls_sizef.add_method('setWidth', None, [('double','w')])
  cls_sizef.add_method('height', 'double', [])
  cls_sizef.add_method('setHeight', None, [('double','h')])

  cls_qrgb = mod.add_class('QRgb')
  cls_qrgb.add_copy_constructor()
  mod.add_container('QVector<QRgb>', retval('QRgb'), 'vector')

  cls_color = mod.add_class('QColor')
  cls_color.add_constructor([('QRgb','col')])
  cls_color.add_constructor([('int','r'), ('int','g'), ('int','b')])
  cls_color.add_constructor([('int','r'), ('int','g'), ('int','b'),('int','a')])
  cls_color.add_copy_constructor()
  cls_color.add_method('rgb', 'QRgb', [])
  cls_color.add_method('rgba', 'QRgb', [])

  cls_qimage = mod.add_class('QImage')
  cls_qimage.add_enum('Format', ('Format_Invalid','Format_Mono','Format_MonoLSB',
    'Format_Indexed8','Format_RGB32','Format_ARGB32',
    'Format_ARGB32_Premultiplied','Format_RGB16',
    'Format_ARGB8565_Premultiplied','Format_RGB666',
    'Format_ARGB6666_Premultiplied','Format_RGB555',
    'Format_ARGB8555_Premultiplied','Format_RGB888','Format_RGB444',
    'Format_ARGB4444_Premultiplied'))
  cls_qimage.add_constructor([])
  cls_qimage.add_constructor([('int','w'), ('int','h'), ('Format','f')])
  cls_qimage.add_copy_constructor()
  cls_qimage.add_method('color', 'QRgb', [('int','i')])
  cls_qimage.add_method('colorTable', 'QVector<QRgb>', [])
  cls_qimage.add_method('fill', None, [('int','color')])
  cls_qimage.add_method('load', 'bool', [('const QString','fileName'),('char*','fmt')])
  cls_qimage.add_method('mirrored', retval('const QImage&'), [('bool','horiz'),('bool','vert')])
  cls_qimage.add_method('width', 'int', [])
  cls_qimage.add_method('height', 'int', [])
  cls_qimage.add_method('setColor', None, [('int','i'),('QRgb','c')])
  cls_qimage.add_method('setPixel', None, [('int','x'),('int','y'),
    ('unsigned int','color')])
  cls_qimage.add_method('setPixel', None, [('int','x'),('int','y'),
    ('QRgb','color')])
  cls_qimage.add_method('setColorTable', None, [('QVector<QRgb>','colors')])
  cls_qpixmap = mod.add_class('QPixmap')
  cls_qpixmap.add_copy_constructor()
  cls_qpixmap.add_method('toImage', retval('const QImage&'), [])
  cls_qpixmap.add_method('fromImage', None, [('const QImage&','image')])
  cls_qpixmap.add_method('convertFromImage', None, [('const QImage&','image')])
  cls_qpixmap.add_method('width', 'int', [])
  cls_qpixmap.add_method('height', 'int', [])

  cls_qwidget = mod.add_class('QWidget')
  cls_qfiledialog = mod.add_class('QFileDialog')
  cls_qfiledialog.add_enum('Option', ('ShowDirsOnly','DontResolveSymlinks','DontConfirmOverwrite','DontUseNativeDialog',
    'ReadOnly','HideNameFilterDetails'))
  cls_qfiledialog.add_method('getOpenFileName', 'QString', [
    param('QWidget*','parent',transfer_ownership=False,null_ok=True),
    ('const QString','caption'),('const QString','dir'),('const QString','filter'),
    param('QString*','selectedFilter',default_value='new QString("")'),
    param('QFlags<QFileDialog::Option>','options', direction=Parameter.DIRECTION_IN, default_value='{}')
    ], is_static=True)

  mod.add_container('QList<QString>', retval('QString'), 'list')
"""
  with open('qtbind.h','w') as fh:
    import pybindgen.typehandlers.codesink as cs
    sink = cs.MemoryCodeSink()
    mod.generate_forward_declarations(sink)
    includes = ['QImage','QFileDialog','QWidget','Qflags']
    print >>fh, '\n'.join(['#include <%s>' % f for f in includes])
    print >>fh, sink.flush()

  with open('qtbind.cpp','w') as fh:
    mod.generate(fh)
"""



#!/usr/bin/env python3
"""
 Python Tiled Plugin
 Copyright 2012-2013, Samuli Tuomola <samuli@tuomola.net>

 This file is part of Tiled.

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the Free
 Software Foundation; either version 2 of the License, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 more details.

 You should have received a copy of the GNU General Public License along with
 this program. If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function
from functools import wraps
from operator import attrgetter, itemgetter
from pybindgen import *
from collections import OrderedDict


class SimpleSortedDict(OrderedDict):
    "naive and ineffecient but adequate for this use"
    def items(self):
        return sorted([[k, self[k]] for k in self], key=itemgetter(0))


def patch_a_prop(func, prop, value_factory):
    """replace an object property after it's given method has executed
    """
    assert callable(value_factory)
    def _decorate(obj, *args, **kwargs):
        ret = func(obj, *args, **kwargs)
        setattr(obj, prop, value_factory())
        return ret

    return wraps(func)(_decorate)


# after a new pybindgen container is instantiated, replace it's methods dictionary
Module.__init__ = patch_a_prop(Module.__init__, 'methods', lambda:SimpleSortedDict())
CppClass.__init__ = patch_a_prop(CppClass.__init__, 'methods', lambda:SimpleSortedDict())


mod = Module('tiled')
mod.functions = SimpleSortedDict()

mod.add_include('"pythonplugin.h"')
mod.add_include('"grouplayer.h"')
mod.add_include('"imagelayer.h"')
mod.add_include('"layer.h"')
mod.add_include('"logginginterface.h"')
mod.add_include('"map.h"')
mod.add_include('"mapobject.h"')
mod.add_include('"objectgroup.h"')
mod.add_include('"tile.h"')
mod.add_include('"tilelayer.h"')
mod.add_include('"tileset.h"')
mod.add_include('"tilesetmanager.h"')

mod.header.writeln('#ifndef _MSC_VER')
mod.header.writeln('#pragma GCC diagnostic ignored "-Wmissing-field-initializers"')
mod.header.writeln('#endif')

# one day PyQt/PySide could be considered
import qtbinding
qtbinding.generate(mod)

#mod.add_include('"qtbind.h"')

tiled = mod.add_cpp_namespace('Tiled')

cls_properties = tiled.add_class('Properties')
cls_properties.add_copy_constructor()
cls_properties.add_method('keys', 'QList<QString>', [])
#cls_propsc = tiled.add_container('QMap<QString,QString>', ('QString','QString'), 'map', cls_properties)

cls_object = tiled.add_class('Object')
cls_object.add_method('properties', retval('Tiled::Properties','p'), [])
cls_object.add_method('propertyAsString', 'QString', [('QString','prop')])
cls_object.add_method('setProperty', None,
    [('QString','prop'),('QString','val')])
cls_object.add_method('setProperty', None,
    [('QString','prop'),('int','val')])
cls_object.add_method('setProperty', None,
    [('QString','prop'),('bool','val')])
cls_object.add_method('propertyType', 'QString', [('QString','prop')])

cls_tile = tiled.add_class('Tile', cls_object)
cls_tile.add_method('id', 'int', [])
cls_tile.add_method('image', retval('const QPixmap&'), [])
cls_tile.add_method('setImage', None, [('const QPixmap&','image')])
cls_tile.add_method('width', 'int', [])
cls_tile.add_method('height', 'int', [])
cls_tile.add_method('size', 'QSize', [])
cls_tile.add_method('type', 'QString', [])

cls_tileset = tiled.add_class('Tileset', cls_object)
cls_sharedtileset = tiled.add_class('SharedTileset')
cls_sharedtileset.add_copy_constructor()
cls_sharedtileset.add_method('data', retval('Tiled::Tileset*',reference_existing_object=True), [])

#cls_tileset.add_enum('Orientation', ('Orthogonal','Isometric'))
cls_tileset.add_method('create', 'Tiled::SharedTileset',
                       [('QString','name'), ('int','tileWidth'), ('int','tileHeight'), ('int','tileSpacing'), ('int','margin')],
                       is_static=True)
cls_tileset.add_method('name', 'QString', [])
cls_tileset.add_method('setName', None, [('QString','name')])
cls_tileset.add_method('fileName', 'QString', [])
cls_tileset.add_method('setFileName', None, [('QString','name')])
cls_tileset.add_method('isExternal', 'bool', [])
cls_tileset.add_method('tileWidth', 'int', [])
cls_tileset.add_method('tileHeight', 'int', [])
cls_tileset.add_method('tileSpacing', 'int', [])
cls_tileset.add_method('margin', 'int', [])
cls_tileset.add_method('tileOffset', 'QPoint', [])
cls_tileset.add_method('setTileOffset', None, [('QPoint','offset')])
#cls_tileset.add_method('orientation', 'Orientation', [])
#cls_tileset.add_method('setOrientation', None, [('Orientation', 'orientation')])
cls_tileset.add_method('gridSize', 'QSize', [])
cls_tileset.add_method('setGridSize', None, [('QSize', 'gridSize')])
cls_tileset.add_method('loadFromImage', 'bool',
    [('const QImage&','img'),('QString','file')])
cls_tileset.add_method('loadImage', 'bool', [])
cls_tileset.add_method('findTile',
    retval('Tiled::Tile*',reference_existing_object=True), [('int','id')])
cls_tileset.add_method('tileAt',
    retval('Tiled::Tile*',reference_existing_object=True), [('int','id')])
cls_tileset.add_method('tileCount', 'int', [])
cls_tileset.add_method('columnCount', 'int', [])
cls_tileset.add_method('rowCount', 'int', [])
cls_tileset.add_method('imageWidth', 'int', [])
cls_tileset.add_method('imageHeight', 'int', [])
cls_tileset.add_method('setTransparentColor', None, [('QColor','col')])
cls_tileset.add_method('transparentColor', 'QColor', [])
cls_tileset.add_method('imageSourceString', 'QString', [])
cls_tileset.add_method('setImageSource', None, [('QString','source')])
cls_tileset.add_method('isCollection', 'bool', [])
cls_tileset.add_method('sharedPointer', 'Tiled::SharedTileset', [])

cls_tile.add_constructor([param('const QPixmap&','image'), param('int','id'),
    param('Tileset*','tileset',transfer_ownership=False)])
cls_tile.add_method('tileset',
    retval('Tiled::Tileset*',reference_existing_object=True), [])

cls_layer = tiled.add_class('Layer', cls_object)

#mod.add_container('QList<Tiled::Tileset*>',
#                retval('Tiled::Tileset*',caller_owns_return=False), 'list')

cls_map = tiled.add_class('Map', cls_object)
cls_map.add_enum('Orientation', ('Unknown','Orthogonal','Isometric','Staggered','Hexagonal'))
cls_map.add_enum('LayerDataFormat', ('XML','Base64','Base64Gzip','Base64Zlib','CSV'))
cls_map.add_enum('RenderOrder', ('RightDown','RightUp','LeftDown','LeftUp'))
cls_map.add_enum('StaggerAxis', ('StaggerX','StaggerY'))
cls_map.add_enum('StaggerIndex', ('StaggerOdd','StaggerEven'))
cls_map.add_constructor([('Orientation','orient'), ('int','w'), ('int','h'),
    ('int','tileW'), ('int','tileH')])
cls_map.add_method('orientation', 'Orientation', [])
cls_map.add_method('setOrientation', None, [('Orientation','orientation')])
cls_map.add_method('renderOrder', 'RenderOrder', [])
cls_map.add_method('setRenderOrder', None, [('RenderOrder','renderOrder')])
cls_map.add_method('width', 'int', [])
cls_map.add_method('setWidth', None, [('int','w')])
cls_map.add_method('height', 'int', [])
cls_map.add_method('setHeight', None, [('int','h')])
cls_map.add_method('tileWidth', 'int', [])
cls_map.add_method('tileHeight', 'int', [])
cls_map.add_method('tileSize', 'QSize', [])
cls_map.add_method('infinite', 'bool', [])
cls_map.add_method('setInfinite', None, [('bool','infinite')])
cls_map.add_method('hexSideLength', 'int', [])
cls_map.add_method('setHexSideLength', None, [('int', 'hexSideLength')])
cls_map.add_method('staggerAxis', 'StaggerAxis', [])
cls_map.add_method('setStaggerAxis', None, [('StaggerAxis','staggerAxis')])
cls_map.add_method('staggerIndex', 'StaggerIndex', [])
cls_map.add_method('setStaggerIndex', None, [('StaggerIndex','staggerIndex')])
cls_map.add_method('layerCount', 'int', [])
cls_map.add_method('tileLayerCount', 'int', [])
cls_map.add_method('objectGroupCount', 'int', [])
cls_map.add_method('imageLayerCount', 'int', [])
cls_map.add_method('groupLayerCount', 'int', [])
cls_map.add_method('backgroundColor', 'QColor', [])
cls_map.add_method('setBackgroundColor', None, [('QColor','col')])
cls_map.add_method('addTileset', None,
    [param('SharedTileset','tileset')])
cls_map.add_method('insertTileset', None,
    [('int','pos'),param('SharedTileset','tileset')])
cls_map.add_method('indexOfTileset', 'int',
    [param('const SharedTileset &','tileset')])
cls_map.add_method('removeTilesetAt', None, [('int','pos')])
cls_map.add_method('replaceTileset', None,
    [param('SharedTileset','oldTileset'),
     param('SharedTileset','newTileset')])
cls_map.add_method('tilesetAt',
    retval('Tiled::SharedTileset'),
    [('int','index')])
cls_map.add_method('tilesetCount', 'int', [])
cls_map.add_method('isTilesetUsed', 'bool',
    [param('const Tileset*','tileset')])
cls_map.add_method('nextLayerId', 'int', [])
cls_map.add_method('setNextLayerId', None, [('int', 'nextLayerId')])
cls_map.add_method('nextObjectId', 'int', [])
cls_map.add_method('setNextObjectId', None, [('int', 'nextObjectId')])

cls_cell = tiled.add_class('Cell')
cls_cell.add_constructor([param('Tiled::Tile*','tile',transfer_ownership=False)])
cls_cell.add_copy_constructor()
cls_cell.add_binary_comparison_operator('==')
cls_cell.add_binary_comparison_operator('!=')
cls_cell.add_method('isEmpty', 'bool', [])
cls_cell.add_instance_attribute('flippedHorizontally', 'bool', getter='flippedHorizontally', setter='setFlippedHorizontally')
cls_cell.add_instance_attribute('flippedVertically', 'bool', getter='flippedVertically', setter='setFlippedVertically')
cls_cell.add_instance_attribute('flippedAntiDiagonally', 'bool', getter='flippedAntiDiagonally', setter='setFlippedAntiDiagonally')
cls_cell.add_instance_attribute('rotatedHexagonal120', 'bool', getter='rotatedHexagonal120', setter='setRotatedHexagonal120')
cls_cell.add_instance_attribute('checked', 'bool', getter='checked', setter='setChecked')
cls_cell.add_method('tile', retval('Tiled::Tile*',reference_existing_object=True), [])
cls_cell.add_method('tileset', retval('Tiled::Tileset*',reference_existing_object=True), [])
cls_cell.add_method('setTile', None, [param('Tiled::Tile*','tile',transfer_ownership=False)])

cls_tilelayer = tiled.add_class('TileLayer', cls_layer)
cls_tilelayer.add_constructor([('QString','name'), ('int','x'), ('int','y'),
    ('int','w'), ('int','h')])
cls_tilelayer.add_method('width', 'int', [])
cls_tilelayer.add_method('height', 'int', [])
cls_tilelayer.add_method('cellAt', retval('Tiled::Cell'),
    [('int','x'),('int','y')])
cls_tilelayer.add_method('setCell', None, [('int','x'),('int','y'),
    ('Cell','c')])
cls_tilelayer.add_method('referencesTileset', 'bool',
    [param('Tileset*','ts',transfer_ownership=False)])
cls_tilelayer.add_method('isEmpty', 'bool', [])

cls_imagelayer = tiled.add_class('ImageLayer', cls_layer)
cls_imagelayer.add_constructor([('QString','name'), ('int','x'), ('int','y')])
cls_imagelayer.add_method('loadFromImage', 'bool',
    [('const QImage&','img'),('QString','file')])
cls_imagelayer.add_method('image', retval('const QPixmap&'), [])
cls_imagelayer.add_method('setImage', None, [('const QPixmap&','image')])

cls_grouplayer = tiled.add_class('GroupLayer', cls_layer)
cls_grouplayer.add_constructor([('QString','name'), ('int','x'), ('int','y')])
cls_grouplayer.add_method('layerCount', 'int', [])
cls_grouplayer.add_method('layerAt', retval('Tiled::Layer*',reference_existing_object=True), [('int','index')])

cls_objectgroup = tiled.add_class('ObjectGroup', cls_layer)

cls_mapobject = tiled.add_class('MapObject', cls_object)
cls_mapobject.add_constructor([])
cls_mapobject.add_constructor([('QString','name'), ('QString','type'),
    ('QPointF','pos'), ('QSizeF','size') ])
cls_mapobject.add_enum('Shape', ('Rectangle','Polygon','Polyline','Ellipse','Text','Point'))
cls_mapobject.add_method('setPosition', None, [('QPointF','pos')])
cls_mapobject.add_method('x', 'double', [])
cls_mapobject.add_method('setX', None, [('double','x')])
cls_mapobject.add_method('y', 'double', [])
cls_mapobject.add_method('setY', None, [('double','y')])
cls_mapobject.add_method('setSize', None, [param('QSizeF','size')])
cls_mapobject.add_method('width', 'double', [])
cls_mapobject.add_method('setWidth', None, [('double','w')])
cls_mapobject.add_method('height', 'double', [])
cls_mapobject.add_method('setHeight', None, [('double','h')])
#cls_mapobject.add_method('setPolygon', None, [param('QPolygonF','pol')])
#cls_mapobject.add_method('polygon', 'QPolygonF', [])
cls_mapobject.add_method('setShape', None, [param('Shape','s')])
cls_mapobject.add_method('shape', 'Shape', [])
#cls_mapobject.add_method('bounds', 'QRectF', [])
cls_mapobject.add_method('setCell', None, [param('const Tiled::Cell','c',)])
cls_mapobject.add_method('cell', retval('const Tiled::Cell'), [])
#cls_mapobject.add_method('setObjectGroup', 'ObjectGroup*', [])
cls_mapobject.add_method('objectGroup', retval('Tiled::ObjectGroup*',reference_existing_object=True), [])
cls_mapobject.add_method('rotation', 'double', [])
cls_mapobject.add_method('setRotation', None, [('double','r')])
cls_mapobject.add_method('isVisible', 'bool', [])
cls_mapobject.add_method('setVisible', None, [('bool','v')])
cls_mapobject.add_method('name', 'QString', [])
cls_mapobject.add_method('setName', None, [('QString','n')])
cls_mapobject.add_method('type', 'QString', [])
cls_mapobject.add_method('setType', None, [('QString','n')])
cls_mapobject.add_method('effectiveType', 'QString', [])

cls_objectgroup.add_constructor([('QString','name'), ('int','x'), ('int','y')])
cls_objectgroup.add_method('addObject', None,
    [param('MapObject*','object',transfer_ownership=True)])
cls_objectgroup.add_method('insertObject', None,
    [('int','index'),param('MapObject*','object',transfer_ownership=False)])
cls_objectgroup.add_method('removeObject', 'int',
    [param('MapObject*','object',transfer_ownership=False)])
cls_objectgroup.add_method('objectAt', retval('Tiled::MapObject*',reference_existing_object=True),[('int','index')])
cls_objectgroup.add_method('objectCount', 'int',[])
#cls_objectgroup.add_method('objectsBoundingRect', 'QRectF', [])
#cls_objectgroup.add_method('usedTilesets', 'QSet<Tileset*>', [])
cls_objectgroup.add_method('referencesTileset', 'bool',
    [param('Tileset*','ts',transfer_ownership=False)])

cls_map.add_method('addLayer', None, [param('TileLayer*','l',transfer_ownership=True)])
cls_map.add_method('addLayer', None, [param('ObjectGroup*','l',transfer_ownership=True)])
cls_map.add_method('addLayer', None, [param('ImageLayer*','l',transfer_ownership=True)])
cls_map.add_method('addLayer', None, [param('GroupLayer*','l',transfer_ownership=True)])

cls_map.add_method('layerAt', retval('Tiled::Layer*',reference_existing_object=True), [('int','index')])

cls_layer.add_method('name', 'QString', [])
cls_layer.add_method('setName', None, [('QString','name')])
cls_layer.add_method('opacity', 'double', [])
cls_layer.add_method('setOpacity', None, [('double','opacity')])
cls_layer.add_method('isVisible', 'bool', [])
cls_layer.add_method('isLocked', 'bool', [])
cls_layer.add_method('isUnlocked', 'bool', [])
cls_layer.add_method('isHidden', 'bool', [])
cls_layer.add_method('setVisible', None, [('bool','visible')])
cls_layer.add_method('setLocked', None, [('bool','locked')])
cls_layer.add_method('map', retval('Tiled::Map*',reference_existing_object=True), [])
cls_layer.add_method('x', 'int', [])
cls_layer.add_method('setX', None, [('int','x')])
cls_layer.add_method('y', 'int', [])
cls_layer.add_method('setY', None, [('int','y')])
cls_layer.add_method('setPosition', None, [('int','x'),('int','y')])
cls_layer.add_method('offset', 'QPointF', [])
cls_layer.add_method('setOffset', None, [('QPointF','offset')])
cls_layer.add_method('isTileLayer', 'bool', [])
cls_layer.add_method('isObjectGroup', 'bool', [])
cls_layer.add_method('isImageLayer', 'bool', [])
cls_layer.add_method('isGroupLayer', 'bool', [])
cls_layer.add_method('asTileLayer', retval('Tiled::TileLayer*',reference_existing_object=True), [])
cls_layer.add_method('asObjectGroup', retval('Tiled::ObjectGroup*',reference_existing_object=True), [])
cls_layer.add_method('asImageLayer', retval('Tiled::ImageLayer*',reference_existing_object=True), [])
cls_layer.add_method('asGroupLayer', retval('Tiled::GroupLayer*',reference_existing_object=True), [])


mod.body.writeln("""
bool isImageLayerAt(Tiled::Map *map, int index) {
    return map->layerAt(index)->isImageLayer();
}
bool isTileLayerAt(Tiled::Map *map, int index) {
    return map->layerAt(index)->isTileLayer();
}
bool isObjectGroupAt(Tiled::Map *map, int index) {
    return map->layerAt(index)->isObjectGroup();
}
Tiled::ImageLayer* imageLayerAt(Tiled::Map *map, int index) {
    return map->layerAt(index)->asImageLayer();
}
Tiled::TileLayer* tileLayerAt(Tiled::Map *map, int index) {
    return map->layerAt(index)->asTileLayer();
}
Tiled::ObjectGroup* objectGroupAt(Tiled::Map *map, int index) {
    return map->layerAt(index)->asObjectGroup();
}
""")

mod.add_function('isImageLayerAt', 'bool',
    [param('Tiled::Map*','map',transfer_ownership=False),('int','index')])
mod.add_function('isTileLayerAt', 'bool',
    [param('Tiled::Map*','map',transfer_ownership=False),('int','index')])
mod.add_function('isObjectGroupAt', 'bool',
    [param('Tiled::Map*','map',transfer_ownership=False),('int','index')])

mod.add_function('imageLayerAt',
    retval('Tiled::ImageLayer*',reference_existing_object=True),
    [param('Tiled::Map*','map',transfer_ownership=False),('int','index')])
mod.add_function('tileLayerAt',
    retval('Tiled::TileLayer*',reference_existing_object=True),
    [param('Tiled::Map*','map',transfer_ownership=False),('int','index')])
mod.add_function('objectGroupAt',
    retval('Tiled::ObjectGroup*',reference_existing_object=True),
    [param('Tiled::Map*','map',transfer_ownership=False),('int','index')])



mod.add_function('loadTilesetFromFile', 'bool',
    [param('Tileset*','ts',transfer_ownership=False),('QString','file')])

mod.body.writeln("""
static bool loadTilesetFromFile(Tiled::Tileset *ts, const QString &file)
{
    QImage img(file);
    return ts->loadFromImage(img, file);
}
""")

mod.add_function('loadTileset', 'Tiled::SharedTileset', [('QString','file')])

mod.body.writeln("""
static Tiled::SharedTileset loadTileset(const QString &file)
{
    return Tiled::TilesetManager::instance()->loadTileset(file);
}
""")

"""
 C++ class PythonScript is seen as Tiled.Plugin from Python script
 (naming describes the opposite side from either perspective)
"""
cls_pp = mod.add_class('PythonScript',
    allow_subclassing=True,
    foreign_cpp_namespace='Python',
    custom_name='Plugin')

"""
 PythonPlugin implements LoggingInterface for messaging to Tiled
"""
cls_logi = tiled.add_class('LoggingInterface', destructor_visibility='private')
cls_logi.add_enum('OutputType', ('INFO','WARNING','ERROR'))
cls_logi.add_method('log', 'void', [('OutputType','type'),('const QString','msg')])


with open('pythonbind.cpp','w') as fh:
    import pybindgen.typehandlers.codesink as cs
    sink = cs.MemoryCodeSink()

    print("""
#ifdef __MINGW32__
#include <cmath> // included before Python.h to fix ::hypot not declared issue
#endif
""", file=fh)

    mod.generate(fh)

    print("""
PyObject* _wrap_convert_c2py__Tiled__LoggingInterface(Tiled::LoggingInterface *cvalue)
{
        PyObject *py_retval;
        PyTiledLoggingInterface *py_LoggingInterface;

        py_LoggingInterface = PyObject_New(PyTiledLoggingInterface, &PyTiledLoggingInterface_Type);
        py_LoggingInterface->flags = PYBINDGEN_WRAPPER_FLAG_NONE;
        py_LoggingInterface->obj = cvalue;
        py_retval = Py_BuildValue((char *) "N", py_LoggingInterface);
        return py_retval;
}

int _wrap_convert_py2c__Tiled__Map___star__(PyObject *value, Tiled::Map * *address)
{
    PyObject *py_retval;
    PyTiledMap *tmp_Map;

    py_retval = Py_BuildValue((char *) "(O)", value);
    if (!PyArg_ParseTuple(py_retval, (char *) "O!", &PyTiledMap_Type, &tmp_Map)) {
        Py_DECREF(py_retval);
        return 0;
    }
    *address = tmp_Map->obj->clone().release();
    Py_DECREF(py_retval);
    return 1;
}
""", file=fh)
    #mod.generate_c_to_python_type_converter(
    #  utils.eval_retval(retval("Tiled::LoggingInterface")),
    #  sink)
    # mod.generate_python_to_c_type_converter(
    #    utils.eval_retval(retval('Tiled::Map*',caller_owns_return=True)),
    #    sink)
    mod.generate_c_to_python_type_converter(
        utils.eval_retval(retval('const Tiled::Map*',reference_existing_object=True)),
        sink)

    print(sink.flush(), file=fh)

# vim: ai ts=4 sts=4 et sw=4 ft=python



"""
Initial support for SNES emulator save states in Tiled
2012, <samuli@tuomola.net>

Note:
- assumes 8x8px tiles in 64x32 map of 4bpp gfx of max 7 palgroups
- only supports bg1 of bgmode 1
http://snesemu.black-ship.net/misc/techdocs/snesdoc.html#GraphicsFormat
"""
from os.path import dirname
from struct import pack,unpack,Struct
from collections import namedtuple
import sys, re, string
import tiled as T


class ZST(T.Plugin):
  @classmethod
  def nameFilter(cls):
    return "zSNES Save State (*.zs?)"

  @classmethod
  def shortName(cls):
    return "zst"

  @classmethod
  def supportsFile(cls, f):
    return open(f).read(26) == 'ZSNES Save State File V0.6'

  @classmethod
  def read(cls, f):
    m = Tiled.Map(Tiled.Map.Orthogonal, 64,32, 8,8)

    # bg1-4 layer bpp counts
    bgmodes = ( (2, 4, 4, 8, 8, 4, 4, 0),
        (2, 4, 4, 4, 2, 2, 0, 0),
        (2, 2, 0, 0, 0, 0, 0, 0),
        (2, 0, 0, 0, 0, 0, 0, 0))

    img = T.qt.QImage(32*8, 32*8, T.qt.QImage.Format_Indexed8)
    #img.fill(255)
    cmap = []
    with open(f, 'rb') as fh:
      #fh.seek(0x66)
      #bgmode = unpack('B', fh.read(1))[0]
      #if bgmode != 7: raise Exception('only mode7 supported atm: '+str(bgmode))

      # supporting new uncompressed save state formats should be just
      # a matter of adding correct addresses for these
      cgrambase = 0x618
      tilemapbase = 0x20C13 # vram:0
      tilebase = 0x2000 # vram offset

      fh.seek(cgrambase)
      cmap = list(parseColors(fh.read(0x200)))
      #img.setColorTable(cmap)

      colors = 16
      tsets = []
      for pal in range(7):
        img.setColorTable(cmap[pal*colors:pal*colors+colors])
        fh.seek(tilemapbase+tilebase)
        readTileset(fh, img)

        tsets.append(Tiled.Tileset.create('Pal%i'%pal, 8,8, 0, 0))
        tsets[pal].data().setTransparentColor(QColor(img.color(0)))
        tsets[pal].data().loadFromImage(img, 'script')

      la = Tiled.TileLayer('Back', 0,0, 64,32)
      fh.seek(tilemapbase)

      for y in range(la.height()):
        for x in range(32):
          t = parseTile(unpack('H', fh.read(2))[0])
          try:
            tile = tsets[t.pal].data().tileAt(t.idx)
            pix = tile.image()
            tile.setImage(pix)
            """ overriding tile gfx could be an alternative to palgroup layers
            img = pix.toImage()
            img.setColorTable(cmap[t.pal*16:t.pal*16+16])
            pix.convertFromImage(img)
            """
            la.setCell(x, y, Tiled.Cell(tile))
          except:
            print('out of range %i,%i: %i' % (x,y,t.idx))

      if la.width() > 32:
        for y in range(la.height()):
          for x in range(32, la.width()):
            t = parseTile(unpack('H', fh.read(2))[0])
            try:
              tile = tsets[t.pal].data().tileAt(t.idx)
              la.setCell(x, y, Tiled.Cell(tile))
            except:
              print('out of range %i,%i: %i' % (x,y,t.idx))

      for pal in range(7):
        m.addTileset(tsets[pal])
      m.addLayer(la)
    return m

def readTileset(dat, img, tvert=32, thoriz=32):
  tw, th = 8, 8
  # iterate tiles
  for i in range(tvert*thoriz):
    tx,ty = i%tvert*tw, i/thoriz*th
    l1 = unpack('>16B', dat.read(16))
    l2 = unpack('>16B', dat.read(16))
    colordat = deplane4bpp([l1,l2])
    # iterate lines in a tile
    for l in range(th):
      # ..pixels in a line
      for p in range(tw):
        c = colordat[l][p]
        img.setPixel(tx+p, ty+l, c)

def deplane2bpp(src, dst):
  "takes 16 bytes, simplify"
  for i in range(8):
    b0 = src[i*2]
    b1 = src[i*2+1]
    if i <= len(dst): dst.append(bytearray(8))
    dst[i][0] += (b0 >> 7) | b1 >> 6 & 2
    dst[i][1] += (b0 >> 6 & 1) | b1 >> 5 & 2
    dst[i][2] += (b0 >> 5 & 1) | b1 >> 4 & 2
    dst[i][3] += (b0 >> 4 & 1) | b1 >> 3 & 2
    dst[i][4] += (b0 >> 3 & 1) | b1 >> 2 & 2
    dst[i][5] += (b0 >> 2 & 1) | b1 >> 1 & 2
    dst[i][6] += (b0 >> 1 & 1) | b1 & 2
    dst[i][7] += (b0 & 1) | b1 & 1 << 1
  return dst

#deplane2bpp( bytearray('3C3C4266 99 66 99 66  99 66 99 3C  42 00 3C 3C'.replace(' ','').decode('hex')) )

def deplane4bpp(src):
  "decode planar 4bpp line of 8px into packed color indexes"
  dst = deplane2bpp(src[1], list())
  for i in range(len(dst)):
    for j in range(len(dst[i])):
      dst[i][j] <<= 2
  return deplane2bpp(src[0], dst)

def parseColors(cgram):
  for i in range(0, len(cgram), 2):
    col = unpack('H', cgram[i:i+2])[0]
    r = (col << 3) & 0xf8
    g = ((col >> 5) << 3) & 0xf8
    b = ((col >> 10) << 3) & 0xf8
    #print (r,g,b),
    yield T.qt.QColor(r,g,b).rgb()

def parseTile(t):
  ret = namedtuple('Tile', 'idx pal prio flipx flipy')

  return ret(idx = t & 1023, pal = t >> 10 & 7, prio = t >> 13,
      flipx = t >> 14 != 0, flipy = t >> 15 != 0)




"""
Mappy support for Tiled
2012-2013, <samuli@tuomola.net>
"""
from base64 import b64encode, b64decode
from collections import OrderedDict
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import os, sys, struct
import pickle

import tiled as T
from lib.mappy_types import BLKSTR, MPHD, fmpchunk


class FMPPicklerMixin:

  @classmethod
  def unpackchunks(cls, f):

    chunks = OrderedDict()
    with open(f, 'rb') as fh:
      frm = fmpchunk()
      frm.unpack(fh)
      frm.data = fh.read(4)
      filelen = frm.len - 16
      chunks[frm.id] = frm
      print(frm)

      while fh.tell() < filelen:
        fc = fmpchunk()
        if not fc.unpack(fh): break
        print(fh.tell(), fc)
        fc.data = fh.read(fc.len)
        chunks[fc.id] = fc

    return chunks

  @classmethod
  def packchunks(cls, fn, chunks):

    with open(fn, 'wb') as fh:
      for k,v in chunks.items():
        print(k, len(v) + v.len) # chunk header + data
        fh.write(v.pack())
        fh.write(v.data)

  @classmethod
  def picklechunks(cls, chunks):
    src = StringIO()
    pi = pickle.Pickler(src, 2)
    pi.dump(chunks)
    print("packlen", src.len)
    return b64encode(src.getvalue())

  @classmethod
  def unpicklechunks(cls, data):
    src = StringIO(b64decode(data))
    print("unpacklen",src.len)
    return pickle.Unpickler(src).load()


class Mappy(T.Plugin, FMPPicklerMixin):

  @classmethod
  def nameFilter(cls):
    return "Mappy (*.fmp)"

  @classmethod
  def shortName(cls):
    return "mappy"

  @classmethod
  def supportsFile(cls, f):
    return open(f,'rb').read(4) == b'FORM'

  @classmethod
  def read(cls, f):

    print('Loading map at',f)
    chunks = cls.unpackchunks(f)
    hd = MPHD()
    # perhaps cpystruct should only read as many bytes as it can handle?
    hd.unpack(chunks['MPHD'].data[:len(hd)])

    m = T.Tiled.Map(T.Tiled.Map.Orthogonal, hd.mapwidth, hd.mapheight,
                    hd.blockwidth, hd.blockheight)
    if hd.type == 2:
      print('Isometric maps not supported at the moment')
      return m

    #TODO: m.setProperty('chunks', cls.picklechunks(chunks))

    tset = T.Tiled.Tileset.create('Tiles', hd.blockwidth, hd.blockheight, 0, 0)
    cmap = list(FMPColormap.unpack(chunks['CMAP'].data))
    tset.data().loadFromImage(FMPTileGfx.unpack(hd, chunks['BGFX'].data, cmap), "")

    blks = FMPBlocks(chunks['BKDT'].data, hd).blocks

    for c in ['LYR'+str(i) for i in range(7,0,-1)]+['BODY']:
      if c not in chunks: continue
      print('populating',c)
      lay = T.Tiled.TileLayer(c,0,0,hd.mapwidth, hd.mapheight)
      lvl = list(FMPLayer.unpack(hd, chunks[c].data))
      FMPLayer.populate(lay, blks, tset.data(), hd, lvl)
      m.addLayer(lay)

    m.addTileset(tset)

    return m


  @classmethod
  def write(cls, m, fn):

    if m.orientation() != T.Tiled.Map.Orthogonal:
      print('Isometric maps not supported at the moment')
      return False
    if not m.property('chunks'):
      raise Exception("Export depends on unparsed binary blobs from original "
                      +"fmp to be stored in the map property 'chunks'")

    print('Writing map at',fn)

    chunks = cls.unpicklechunks(m.property('chunks'))
    hd = MPHD()
    hd.unpack(chunks['MPHD'].data[:len(hd)])
    blks = FMPBlocks(chunks['BKDT'].data, hd).blocks

    for i in range(m.layerCount()):

      if not isTileLayerAt(m, i): continue
      l = tileLayerAt(m, i)

      chunks[l.name()].data = FMPLayer.pack(hd, blks, l, i)

    cls.packchunks(fn, chunks)

    return True


class FMPBlocks:

  @classmethod
  def unpack(cls, chunk, hd):
    mod = hd.blockwidth * hd.blockheight * ((hd.blockdepth+1)/8)
    for pos in range(0, hd.numblockstr*hd.blockstrsize, hd.blockstrsize):
      b = BLKSTR()
      b.unpack(chunk[pos:pos+hd.blockstrsize])
      if hd.type == 0:
        for i in range(4):
          b.olay[i] /= mod
      yield b

  def __init__(self, chunk, hd):
    self.blocks = list(FMPBlocks.unpack(chunk, hd))


class FMPTileGfx:

  @classmethod
  def unpack(cls, hd, dat, cmap):
    n = 0
    w,h = hd.blockwidth*10, int(hd.blockheight*hd.numblockgfx/10+hd.blockheight)
    fmt = T.qt.QImage.Format_Indexed8 if hd.blockdepth==8 else T.qt.QImage.Format_ARGB32
    img = T.qt.QImage(w, h, fmt)
    img.setColorTable(cmap)

    for i in range(hd.numblockgfx):
      col,row = int(i%10), int(i/10)
      tx,ty = col*hd.blockwidth, row*hd.blockheight

      for y in range(hd.blockheight):
        for x in range(hd.blockwidth):
          c = struct.unpack('B', dat[n:n+1])[0]
          if hd.blockdepth==8:
            img.setPixel(tx+x, ty+y, c)
          else:
            img.setPixel(tx+x, ty+y, cmap[c])
          n+=1

    return img


class FMPLayer:

  @classmethod
  def unpack(cls, hd, dat):
    for i in range(0,hd.mapheight*hd.mapwidth*2,2):
      d = struct.unpack('<H', dat[i:i+2])[0]
      #d<0 is anim?
      if hd.type == 0: d /= hd.blockstrsize

      yield d

  @classmethod
  def pack(cls, hd, blocks, layer, laynum):
    dat = StringIO()
    print(hd)
    for y in range(layer.height()):
      for x in range(layer.width()):
        cell = layer.cellAt(x, y)
        tid = 0 if cell.isEmpty() else cell.tile.id()
        n = blocks[tid].olay[laynum]
        if hd.type == 0:
          n *= hd.blockstrsize
        dat.write( struct.pack('<H', n) )

    return dat.getvalue()

  @classmethod
  def populate(cls, layer, blocks, tileset, hd, ldata):

    for fg in range(4):
      i = 0

      for y in range(hd.mapheight):
        for x in range(hd.mapwidth):
          v = int(ldata[i])

          if v >= len(blocks):
            pass#print 'unknown block at',i
          else:
            n = int(blocks[v].olay[fg])
            if n != 0:
              ti = tileset.tileAt(n)
              if ti is None:
                print('invalid tile',n,'at',x,y)
              else:
                layer.setCell(x, y, T.Tiled.Cell(ti))
          i += 1


class FMPColormap:

  @classmethod
  def unpack(self, cmap):
    """cmap -- rgb bytearray"""
    print('got',len(cmap))
    for i in range(0,len(cmap),3):
      r,g,b = struct.unpack('3B', cmap[i:i+3])
      yield T.qt.QColor(r,g,b).rgb()

  @classmethod
  def pack(self, cmap):
    """cmap -- T.qt.QImage.colorTable"""
    yield fmpchunk(id='CMAP', len=len(list(cmap))).pack()
    for c in cmap:
      yield struct.pack('3B', (c.qRed,c.qGreen,c.qBlue))





from .cpystruct import *

"""
#define AN_END -1     /* Animation types, AN_END = end of anims */
#define AN_NONE 0     /* No anim defined */
#define AN_LOOPF 1    /* Loops from start to end, then jumps to start etc */
#define AN_LOOPR 2    /* As above, but from end to start */
#define AN_ONCE 3     /* Only plays once */
#define AN_ONCEH 4    /* Only plays once, but holds end frame */
#define AN_PPFF 5     /* Ping Pong start-end-start-end-start etc */
#define AN_PPRR 6     /* Ping Pong end-start-end-start-end etc */
#define AN_PPRF 7     /* Used internally by playback */
#define AN_PPFR 8     /* Used internally by playback */
#define AN_ONCES 9    /* Used internally by playback */
"""

class BLKSTR(CpyStruct("""
int olay[4];
unsigned int user1;
unsigned int user2; // user long data */
unsigned short user3;
unsigned short user4;  // user short data */
unsigned char user5;
unsigned char user6;
unsigned char user7;  // user byte data */
unsigned char tl; // bits for collision detection */
""")): pass
"""for newer fmp version?
unsigned char tr;
unsigned char bl;
unsigned char br;
unsigned char trigger;  // bit to trigger an event */
unsigned char unused1;
unsigned char unused2;
unsigned char unused3;
"""

class ANISTR(CpyStruct("""
signed char antype; // Type of anim, AN_? */
signed char andelay;  // Frames to go before next frame */
signed char ancount;  // Counter, decs each frame, till 0, then resets to andelay */
signed char anuser; // User info */
long ancuroff;  // Points to current offset in list */
long anstartoff;  // Points to start of blkstr offsets list, AFTER ref. blkstr offset */
long anendoff;
""")): pass

class OBJSTR(CpyStruct("""
int xpos; int ypos; // pixel position in map to handle */
int gfxid; int tileid;
int gxoff; int gyoff; // offset into graphic */
int gwidth; int gheight;
int ghandlexoff; int ghandleyoff; // handle pos, from gxoff, gyoff */
int show; // display mode */
int user1; int user2; int user3; int user4; int user5; int user6; int user7;
int flags;
""")): pass

class GENHEAD(CpyStruct("""
char id1; char id2; char id3; char id4;  // 4 byte header id. */
long headsize;    // size of header chunk. */
""")): pass

#// char M,P,H,D;  4 byte chunk identification. */
#// long int mphdsize; size of map header. */
class MPHD(CpyStruct("""
BYTE mapverhigh;  // map version number to left of . (ie X.0). */
BYTE mapverlow;   // map version number to right of . (ie 0.X). */
BYTE lsb;   // if 1, data stored LSB first, otherwise MSB first. */
BYTE type;  // 0=old, 1=new, 2=iso
short mapwidth; // width in blocks. */
short mapheight;  // height in blocks. */
short reserved1;
short reserved2;
short blockwidth; // width of a block (tile) in pixels. */
short blockheight;  // height of a block (tile) in pixels. */
short blockdepth; // depth of a block (tile) in planes (ie. 256 colours is 8) */
short blockstrsize; // size of a block data structure */
short numblockstr;  // Number of block structures in BKDT */
short numblockgfx;  // Number of 'blocks' in graphics (BODY) */
BYTE skip;
SBYTE trans8bit;
short transhi;
""")): pass

#// char E,D,H,D;  4 byte chunk identification. */
#// long int edhdsize; size of editor header. */
class EDHD(CpyStruct("""
short  xmapoffset; // editor offset, in blocks, from left. */
short  ymapoffset; // editor offset, in blocks, from right. */
long fgcolour;  // fg colour for text, buttons etc. */
long bgcolour;  // bg colour for text, buttons etc. */
short swidth;   // width of current screen res */
short sheight;  // height of current screen res */
short strtstr;  // first structure in view */
short strtblk;  // first block graphic in view */
short curstr;   // current block structure */
short curanim;  // current anim structure */
short animspd;  // gap in frames between anims */
short span;   // control panel height */
short numbrushes; // number of brushes to follow. */
""")): pass

class MAPFILL(CpyStruct("short leftedge, width, yoff;")): pass

class fmpchunk(CpyStruct(":4s id; :I len;",True)): pass

class BODY(CpyStruct(":H dat[len]", True)): pass




"""
  CreepyStruct - Convenience class for (un)packing structured binaries
  (c)2011-2012, <samuli@tuomola.net>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import re, sys, struct

# struct.pack format characters prefixed by :
REFMT = r':[@!<>=]?[0-9xcbBhHiIlLqQfdspP]+'
# possible array definition or value assignment
REARR = r'(?:\[(\w+)\])?(?:\s*=\s*([0-9a-fx]+))?'
# skip C or python style comments to the end of line
RECMT = r'\s*(?:#.*|//.*)?'
# whole line with at least format/struct ref and attribute name
REPCK = r'(%s|[\s\w]+|,)\s+(\w+)%s%s[;]?%s' % (REFMT, RECMT, REARR, RECMT)

FORMATS = {
  'c':'char',
  'b':'signed char|SBYTE',
  'B':'unsigned char|uchar|BYTE',
  '?':'_Bool|bool',
  'h':'short|SWORD',
  'H':'unsigned short|ushort|UWORD',
  'i':'int',
  'I':'unsigned int|uint|WORD',
  'l':'long',
  'L':'unsigned long|ulong|DWORD',
  'q':'long long',
  'Q':'unsigned long long',
  'f':'float',
  'd':'double'
}
fdict = {}
for f in FORMATS:
  fdict.update( [(p,f) for p in FORMATS[f].split('|')] )

class CpySkeleton(struct.Struct):
  """ Not to be used directly, use CpyStruct() to build a class """

  def __init__(self, dat=None, **kws):
    """ Takes keyword arguments to initialize attributes """
    struct.Struct.__init__(self, getattr(self, '__fstr'))

    self.validate()

    if len(kws) > 0:
      for k in kws:
        setattr(self, k, kws[k])

    if dat != None:
      self.unpack(dat)

  def validate(self):
    "check that the extending classes specify valid struct"

    for i,(f,n,a,v) in enumerate(self.formats):
      if a != '' and not a.isdigit() and i < len(self.formats):
        for b in self.formats[i:]:
          if a=='' or b[2].isdigit():
            raise Exception('Varlength arrays only '+
              'supported at end of struct: %s[%s]' % (n,a))

  def pack(self):
    "convert member values to binary"
    ret = ''

    for i,(f,n,a,v) in enumerate(self.formats):
      v = getattr(self, n)
      if issubclass(v.__class__, CpySkeleton):
        ret += v.pack()
        continue

      if type(f) != type(struct.Struct) and f in fdict:
        f = fdict[f]

      if a != '':
        # array (possibly varlength)
        a = int(a) if a.isdigit() else getattr(self, a)
        if f=='c':
          # chararray as string
          f = str(a)+'s'
          a = 0

      if type(f) == type(struct.Struct):
        fstr = getattr(f, '__fstr')
        # temp hack because of qmap<str,str> limitations in tiled map props
        sz = struct.calcsize(fstr)
        m = re.match('<[0-9]*[sc]',fstr)
        if a != '' and a > 0 and (m is None or m.group(0) != fstr):
          for j in range(len(v)): #a):
            v2 = f(v[j].ljust(sz,'\0'))
            ret += v2.pack()
        elif type(v) == list:
          print('pack list') # todo
        else:
          #v = f(v.ljust(sz,'\0'))
          ret += struct.pack(getattr(f, '__fstr'), v)
      else:
        f = getattr(self, '__endianflag') + f
        if type(v) is list:
          for x in v:
            ret += struct.pack(f, x)
        else:
          # just an ordinary var
          ret += struct.pack(f, v)

    return ret

  def unpack(self, dat):
    """
    Takes a string, file, mmap or StringIO instance
    """
    rawpos = 0  # position in binary for custom types
    pos = 0  # in case substruct handles multiple values

    # readable, e.g. file, mmap, StringIO, BufferedReader
    if hasattr(dat,'read') and callable(dat.read):
      # this doesn't cover varlen members, for that dat is read directly
      buf = dat.read(len(self))
    else:
      buf = dat

    unpacked = None

    if hasattr(self, 'fromraw'):
      sz = struct.calcsize(getattr(self, '__fstr'))
      unpacked = self.fromraw(buf[:sz])

    if unpacked is None:
      unpacked = list(struct.Struct.unpack(self, buf))

    if hasattr(self, 'fromval'):
      unpacked = self.fromval(unpacked)

    if not hasattr(unpacked, '__iter__'):
      unpacked = [unpacked]

    #if len(self.formats) != len(unpacked):
    #  print unpacked
    #  raise LookupError('Unmatched number of unpacked variables: %i != %i'
    #    % (len(self.formats),len(unpacked)))

    for i,(f,n,a,v) in enumerate(self.formats):
      if a != '' and not a.isdigit(): break

      arlen = 0
      if a.isdigit() and (type(f) is type or fdict[f] != 'c'):
        arlen = int(a)

      v = unpacked[pos]

      if type(f) == type(struct.Struct):
        sz = struct.calcsize(getattr(f, '__fstr'))
        if arlen > 0:
          arr = []
          for i in range(arlen):
            arr.append( f(buf[rawpos:rawpos+sz]) )
            rawpos += sz
            pos += len(f.formats)
          setattr(self, n, arr)
        else:
          setattr(self, n, f(buf[rawpos:rawpos+sz]))
          rawpos += sz
          pos += len(f.formats)

      else:
        if arlen > 0:
          setattr(self, n, unpacked[pos:pos+arlen])
          pos += arlen
        else:
          #print(f,n,a,v)
          if f.endswith('s'): v = v.decode()  # strings for py3
          setattr(self, n, v)
          pos += 1

        rawpos += getattr(self, "__fsz")[i]

    # variable-length members
    for f,n,a,v in self.formats:
      if a != '' and not a.isdigit():
        c = getattr(self, a)

        # one level of custom varlen types should be enough for everyone
        if isinstance(c, struct.Struct):
          # would it be better to delegate this to the custom class?
          c = getattr(c, c.__slots__[0])

        #dat.seek(rawpos)
        if type(f) == type(struct.Struct):
          # custom types
          sz = struct.calcsize(getattr(f, '__fstr'))
          arr = []
          for i in range(c):
            arr.append( f(dat.read(sz)) )
            rawpos += sz
          setattr(self, n, arr)
        else:
          # primitives
          f = str(c)+'s' if fdict[f]=='c' else str(c)+fdict[f]
          sz = struct.calcsize(f)
          val = struct.unpack(f, dat.read(sz))
          #if len(val) == 1: val = val[0]
          rawpos += sz
          setattr(self, n, val)

    return buf

  def __len__(self):
    return struct.calcsize(getattr(self, '__fstr'))

  def __str__(self):
    ret = self.__class__.__name__+'['
    for f,n,a,v in self.formats:
      ret += '%s=%s,' % (n,getattr(self, n, ''))
    return ret[:-1]+']'


def peek(s, n):
  p = s.tell()
  r = s.read(n)
  s.seek(p)
  return r

def parseformat(fmt, callscope=None):
  fstr = ''
  sz = []
  for i,(f,n,a,v) in enumerate(fmt):
    if f == ',' and i > 0:
      fmt[i] = (fmt[i-1][0],n,a,v)
      f = fmt[i-1][0]
    elif f == ',':
      raise Exception('Unexpected comma at '+str(fmt[i]))

    fs = ''
    if f in fdict:
      if a.isdigit():
        fs = 's' if fdict[f] == 'c' else fdict[f]
      elif a != '' and not a.isdigit():
        # varlength array, read separately
        pass
      else:
        # C type
        fs = fdict[f]
    elif type(f) is type(CpySkeleton):
      # might have an alignment issue when mixing endians..
      fs = re.sub('[<>]?','',f.__fstr)
    elif f in callscope.f_globals:
      # resolve references to other CpyStructs
      fmt[i] = (callscope.f_globals[f],n,a,v)
      fs = re.sub('[<>]?','',fmt[i][0].__fstr)
    elif f[0] == ':':
      # struct format uses colon as prefix for explicitness
      fs = f[1:]
      fmt[i] = (f[1:],n,a,v)
    else:
      raise Exception('Unknown format at '+str(fmt[i]))

    if a.isdigit():
      # in case it's e.g. custom type of a string
      if fs[0].isdigit():
        fs *= int(a)
      else:
        fs = a + fs
    #elif a != '':
    #  fstr += '{'+a+'}'

    sz.append(struct.calcsize(fs))
    fstr += fs

  return (fmt, fstr, sz)

def CpyStruct(s, endianflg='<'):
  """ Call with a string specifying
  C-like struct to get a Struct class """
  # f=format, n=name, a=arraydef, v=default value
  fmt = [(f.strip(),n,a,v) for f,n,a,v in re.findall(REPCK, s)]
  # peek into caller's namespace in case they refer to custom classes
  callscope = sys._getframe(1)
  try:
    fmt, fstr, fsz = parseformat(fmt, callscope)
  finally:
    del callscope

  # for backwards compatibility
  if endianflg == True:
    endianflg = '>'
  elif endianflg == False:
    endianflg = '<'

  d = {}
  d['__endianflag'] = endianflg
  d['__fstr'] = endianflg + fstr
  d['__fsz'] = fsz
  d['__slots__'] = [n for f,n,a,v in fmt]
  d['formats'] = fmt

  for f,n,a,v in fmt:
    if v != '': d[n] = int(v,0)

  return type('', (CpySkeleton,), d)

